package org.rosuda.jrs;

import java.io.IOException;
import java.util.Map;

import javax.script.Bindings;
import javax.script.ScriptContext;
import javax.script.ScriptException;

import org.rosuda.REngine.REXP;
import org.rosuda.REngine.REXPMismatchException;
import org.rosuda.REngine.REngine;
import org.rosuda.REngine.REngineCallbacks;
import org.rosuda.REngine.REngineException;
import org.rosuda.REngine.REngineOutputInterface;

/**
 * R implementation of ScriptEngine using REngine (JRI or Rserve).
 * @author Martyn Tebby
 */
public class RScriptEngine extends BaseScriptEngine implements ScriptEngineEx,
		REngineCallbacks, REngineOutputInterface { //, REngineInputInterface {

	/** Underlying R implementation. */
 	protected REngine engine;
 	private boolean closeREPL;

 	/**
 	 * Create the ScriptEngine with the given R implementation.
 	 * @param engine The REngine implementation.
 	 */
	public RScriptEngine(REngine engine) {
		setEngine(engine);
	}

	protected RScriptEngine(boolean closeREPL) {
		this.closeREPL = closeREPL;
	}

	protected void setEngine(REngine engine) {
		this.engine = engine;
		try {
			// setup the cache of Opaque objects that maybe passed to and fro
			myEval(ROpaque.name + " = list()");
		}
		catch(Exception e) {
			throw new RuntimeException(e);
		}	
	}

	@Override
	public Object eval(String script, ScriptContext context) throws ScriptException {
		if(context != null) {
			Bindings bindings = context.getBindings(ScriptContext.ENGINE_SCOPE);
			if(bindings != null) {
				for(Map.Entry<String, Object> entry : bindings.entrySet()) {
					assignArg(entry.getKey(), entry.getValue());
				}
			}
		}
		try {
			REXP rexp = myEval(script);
			return RexpConvert.rexp2jobj(rexp);
		}
		catch(Exception rme) {
			throw new ScriptException(rme);
		}
	}

	@Override
	@SuppressWarnings("unchecked")
	public <T> T call(Class<T> returnType, Object obj,
			String function, Object... args) throws ScriptException {
		Map<String, Object> namedArgs = context.getBindings(ScriptContext.ENGINE_SCOPE);
		String cmd = setupArgs(obj, function, args, namedArgs);
		if(returnType == void.class) {
			cmd += "NULL;";	// don't process return values when not needed
		}
		ROpaque oo = null;
		try {
			if(returnType != null && ROpaque.class.isAssignableFrom(returnType)) {
				oo = (ROpaque) returnType.newInstance();
				cmd = oo + "=" + cmd + "NULL;";
			}
			REXP rret = myEval(cmd);
			Object ret = oo != null ? oo : RexpConvert.rexp2jobj(rret, returnType);
			return (T) ret;
		}
		catch(Exception e) {
			throw new ScriptException(e);
		}
	}

	@Override
	public void close() {
		try {
			if(closeREPL) {
				myEval("q('no', 0, FALSE)");
			}
		}
		catch(Exception e) {
			e.printStackTrace();
		}
		engine.close();
	}

	/** Internal eval method. */
	private REXP myEval(String cmd) throws REngineException, REXPMismatchException {
		return engine.parseAndEval(cmd);
	}

	/** Return a string representing the function call. */
	private String setupArgs(Object obj, String function, Object[] args, Map<String, Object> namedArgs) {
		StringBuilder cmd = new StringBuilder();
		cmd.append(function).append("(");
		int pos = 0;
		if(obj != null) {
			writeArg(null, obj, pos++, cmd);
		}
		if(args != null) {
			for(Object arg : args) {
				writeArg(null, arg, pos++, cmd);
			}
		}
		if(namedArgs != null) {
			for(Map.Entry<String, Object> entry : namedArgs.entrySet()) {
				writeArg(entry.getKey(), entry.getValue(), pos++, cmd);
			}
			namedArgs.clear();
		}
		cmd.append(");");
		return cmd.toString();
	}

	/** Add the argument representation to the passed StringBuilder. */
	private void writeArg(String name, Object value, int pos, StringBuilder cmd) {
		cmd.append(pos > 0 ? "," : "");
		cmd.append(name != null ? name + "=" : "");
		cmd.append(assignArg("arg" + pos, value));
	}

	/** Return a string representation of the function argument and other setup. */
	private String assignArg(String name, Object value) {
		if(value == null) {
			name = "NULL";
		}
		else if(value instanceof ROpaque) {
			name = value.toString();
		}
		else {
			try {
				engine.assign(name, RexpConvert.jobj2rexp(value), null);
			}
			catch(Exception e) {
				throw new RuntimeException(e);
			}
		}
		return name;
	}

	// REngineOutputInterface methods

	@Override
	public void RFlushConsole(REngine eng) {
		try {
			getContext().getWriter().flush();
		} catch(IOException ioe) {
			throw new RuntimeException(ioe);
		}
	}

	@Override
	public void RShowMessage(REngine eng, String msg) {
		try {
			getContext().getErrorWriter().write(msg);
		} catch(IOException ioe) {
			throw new RuntimeException(ioe);
		}
	}

	@Override
	public void RWriteConsole(REngine eng, String msg, int otype) {
		try {
			(otype == 0 ? getContext().getWriter() : getContext().getErrorWriter()).write(msg);
		} catch(IOException ioe) {
			throw new RuntimeException(ioe);
		}
	}

//	@Override
	// Does not seam to work
	public String RReadConsole(REngine eng, String prompt, int addToHistory) {
		try {
			char[] cbuf = new char[1024];
			getContext().getErrorWriter().write(prompt + ": ");
			getContext().getErrorWriter().flush();
			getContext().getReader().read(cbuf);
			return new String(cbuf);
		} catch(IOException ioe) {
			throw new RuntimeException(ioe);
		}
	}

}
